'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var web3_js = require('@solana/web3.js');
var splToken = require('@solana/spl-token');
var BN = require('bn.js');
var SdkError = require('../../errors/SdkError.js');
var EditionMarkerAccount = require('../../programs/tokenMetadata/accounts/EditionMarkerAccount.js');
var PrintEditionAccount = require('../../programs/tokenMetadata/accounts/PrintEditionAccount.js');
var mintNewEditionFromMasterEditionViaVaultProxyBuilder = require('../../programs/tokenMetadata/transactionBuilders/mintNewEditionFromMasterEditionViaVaultProxyBuilder.js');
var mintNewEditionFromMasterEditionViaTokenBuilder = require('../../programs/tokenMetadata/transactionBuilders/mintNewEditionFromMasterEditionViaTokenBuilder.js');
var Operation = require('../../types/Operation.js');
var MetadataAccount = require('../../programs/tokenMetadata/accounts/MetadataAccount.js');
var OriginalEditionAccount = require('../../programs/tokenMetadata/accounts/OriginalEditionAccount.js');
var TransactionBuilder = require('../../utils/TransactionBuilder.js');
var createMintAndMintToAssociatedTokenBuilder = require('../../programs/token/transactionBuilders/createMintAndMintToAssociatedTokenBuilder.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var BN__default = /*#__PURE__*/_interopDefaultLegacy(BN);

const Key = 'PrintNewEditionOperation';
const printNewEditionOperation = Operation.useOperation(Key);
const printNewEditionOperationHandler = {
  handle: async (operation, metaplex) => {
    const {
      originalMint,
      newMint = web3_js.Keypair.generate(),
      newMintAuthority = metaplex.identity(),
      newUpdateAuthority = newMintAuthority.publicKey,
      newOwner = newMintAuthority.publicKey,
      newFreezeAuthority,
      payer = metaplex.identity(),
      tokenProgram,
      associatedTokenProgram,
      confirmOptions
    } = operation.input; // Original NFT.

    const originalMetadata = await MetadataAccount.MetadataAccount.pda(originalMint);
    const originalEdition = await OriginalEditionAccount.OriginalEditionAccount.pda(originalMint);
    const originalEditionAccount = OriginalEditionAccount.OriginalEditionAccount.fromMaybe(await metaplex.rpc().getAccount(originalEdition));

    if (!originalEditionAccount.exists) {
      throw new SdkError.AccountNotFoundError(originalEdition, 'OriginalEdition', `Ensure the provided mint address for the original NFT [${originalMint.toBase58()}] ` + `is correct and that it has an associated OriginalEdition PDA.`);
    }

    const edition = new BN__default["default"](originalEditionAccount.data.supply, 'le').add(new BN__default["default"](1));
    const originalEditionMarkPda = await EditionMarkerAccount.EditionMarkerAccount.pda(originalMint, edition); // New NFT.

    const newMetadata = await MetadataAccount.MetadataAccount.pda(newMint.publicKey);
    const newEdition = await PrintEditionAccount.PrintEditionAccount.pda(newMint.publicKey);
    const lamports = await splToken.getMinimumBalanceForRentExemptMint(metaplex.connection);
    const newAssociatedToken = await splToken.getAssociatedTokenAddress(newMint.publicKey, newOwner, false, tokenProgram, associatedTokenProgram);
    const sharedInput = {
      lamports,
      edition,
      newMint,
      newMetadata,
      newEdition,
      newMintAuthority,
      newUpdateAuthority,
      newOwner,
      newAssociatedToken,
      newFreezeAuthority,
      payer,
      originalMetadata,
      originalEdition,
      originalEditionMarkPda,
      tokenProgram,
      associatedTokenProgram
    };
    let transactionBuilder;

    if (operation.input.via === 'vault') {
      transactionBuilder = printNewEditionBuilder({
        via: 'vault',
        vaultAuthority: operation.input.vaultAuthority,
        safetyDepositStore: operation.input.safetyDepositStore,
        safetyDepositBox: operation.input.safetyDepositBox,
        vault: operation.input.vault,
        tokenVaultProgram: operation.input.tokenVaultProgram,
        ...sharedInput
      });
    } else {
      var _operation$input$orig, _operation$input$orig2;

      const originalTokenAccountOwner = (_operation$input$orig = operation.input.originalTokenAccountOwner) !== null && _operation$input$orig !== void 0 ? _operation$input$orig : metaplex.identity();
      const originalTokenAccount = (_operation$input$orig2 = operation.input.originalTokenAccount) !== null && _operation$input$orig2 !== void 0 ? _operation$input$orig2 : await splToken.getAssociatedTokenAddress(originalMint, originalTokenAccountOwner.publicKey, false, tokenProgram, associatedTokenProgram);
      transactionBuilder = printNewEditionBuilder({
        via: 'token',
        originalTokenAccountOwner,
        originalTokenAccount,
        ...sharedInput
      });
    }

    const {
      signature
    } = await metaplex.rpc().sendAndConfirmTransaction(transactionBuilder, undefined, confirmOptions);
    return {
      mint: newMint,
      metadata: newMetadata,
      edition: newEdition,
      associatedToken: newAssociatedToken,
      transactionId: signature
    };
  }
};
const printNewEditionBuilder = params => {
  const {
    // Data.
    lamports,
    edition,
    // New NFT.
    newMint,
    newMetadata,
    newEdition,
    newMintAuthority,
    newUpdateAuthority,
    newOwner,
    newAssociatedToken,
    newFreezeAuthority,
    payer,
    // Master NFT.
    originalMetadata,
    originalEdition,
    originalEditionMarkPda,
    // Programs.
    tokenProgram,
    associatedTokenProgram,
    // Instruction keys.
    createAccountInstructionKey,
    initializeMintInstructionKey,
    createAssociatedTokenInstructionKey,
    mintToInstructionKey,
    printNewEditionInstructionKey = 'printNewEdition'
  } = params;
  let printNewEditionViaBuilder;

  if (params.via === 'vault') {
    printNewEditionViaBuilder = mintNewEditionFromMasterEditionViaVaultProxyBuilder.mintNewEditionFromMasterEditionViaVaultProxyBuilder({
      edition,
      newMetadata,
      newEdition,
      masterEdition: originalEdition,
      newMint,
      editionMarkPda: originalEditionMarkPda,
      newMintAuthority,
      payer,
      vaultAuthority: params.vaultAuthority,
      safetyDepositStore: params.safetyDepositStore,
      safetyDepositBox: params.safetyDepositBox,
      vault: params.vault,
      newMetadataUpdateAuthority: newUpdateAuthority,
      metadata: originalMetadata,
      tokenVaultProgram: params.tokenVaultProgram,
      instructionKey: printNewEditionInstructionKey
    });
  } else {
    printNewEditionViaBuilder = mintNewEditionFromMasterEditionViaTokenBuilder.mintNewEditionFromMasterEditionViaTokenBuilder({
      edition,
      newMetadata,
      newEdition,
      masterEdition: originalEdition,
      newMint,
      editionMarkPda: originalEditionMarkPda,
      newMintAuthority,
      payer,
      tokenAccountOwner: params.originalTokenAccountOwner,
      tokenAccount: params.originalTokenAccount,
      newMetadataUpdateAuthority: newUpdateAuthority,
      metadata: originalMetadata,
      instructionKey: printNewEditionInstructionKey
    });
  }

  return TransactionBuilder.TransactionBuilder.make() // Create the mint account and send one token to the holder.
  .add(createMintAndMintToAssociatedTokenBuilder.createMintAndMintToAssociatedTokenBuilder({
    lamports,
    decimals: 0,
    amount: 1,
    createAssociatedToken: true,
    mint: newMint,
    payer,
    mintAuthority: newMintAuthority,
    owner: newOwner,
    associatedToken: newAssociatedToken,
    freezeAuthority: newFreezeAuthority,
    tokenProgram,
    associatedTokenProgram,
    createAccountInstructionKey,
    initializeMintInstructionKey,
    createAssociatedTokenInstructionKey,
    mintToInstructionKey
  })) // Mint new edition.
  .add(printNewEditionViaBuilder);
};

exports.printNewEditionBuilder = printNewEditionBuilder;
exports.printNewEditionOperation = printNewEditionOperation;
exports.printNewEditionOperationHandler = printNewEditionOperationHandler;
//# sourceMappingURL=printNewEdition.js.map
