'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var EventEmitter = require('eventemitter3');
var useDisposable = require('./useDisposable.js');
var SdkError = require('../errors/SdkError.js');

const useTask = callback => {
  // State.
  let status = 'pending';
  let result = undefined;
  let error = undefined;
  const eventEmitter = new EventEmitter.EventEmitter(); // Getters.

  const getStatus = () => status;

  const getResult = () => result;

  const getError = () => error;

  const isPending = () => status === 'pending';

  const isRunning = () => status === 'running';

  const isCompleted = () => status !== 'pending' && status !== 'running';

  const isSuccessful = () => status === 'successful';

  const isFailed = () => status === 'failed';

  const isCanceled = () => status === 'canceled'; // Setters.


  const setStatus = newStatus => {
    if (status === newStatus) return;
    status = newStatus;
    eventEmitter.emit('statusChange', newStatus);
  }; // Run methods.


  const forceRun = async (options = {}) => {
    const disposable = useDisposable.useDisposable(options.signal).onCancel(cancelError => {
      setStatus('canceled');
      error = cancelError;
    });
    return disposable.run(async scope => {
      const {
        isCanceled,
        throwIfCanceled
      } = scope;

      try {
        // Start loading.
        setStatus('running');
        result = undefined;
        error = undefined;
        result = await Promise.resolve(callback(scope));
        throwIfCanceled();
        setStatus('successful'); // Return the loaded result.

        return result;
      } catch (newError) {
        // Capture the error and reset the result.
        error = newError;
        result = undefined;
        setStatus(isCanceled() ? 'canceled' : 'failed'); // Re-throw the error.

        throw error;
      }
    });
  };

  const run = async (options = {}) => {
    var _options$force;

    if (isRunning()) {
      throw new SdkError.TaskIsAlreadyRunningError();
    }

    if (isPending() || ((_options$force = options.force) !== null && _options$force !== void 0 ? _options$force : false)) {
      return forceRun(options);
    }

    if (isSuccessful()) {
      return getResult();
    }

    throw getError();
  };

  return {
    getStatus,
    getResult,
    getError,
    isPending,
    isRunning,
    isCompleted,
    isSuccessful,
    isFailed,
    isCanceled,
    run,

    loadWith(preloadedResult) {
      setStatus('successful');
      result = preloadedResult;
      error = undefined;
      return this;
    },

    reset() {
      setStatus('pending');
      result = undefined;
      error = undefined;
      return this;
    },

    onStatusChange(callback) {
      eventEmitter.on('statusChange', callback);
      return this;
    },

    onStatusChangeTo(status, callback) {
      return this.onStatusChange(newStatus => status === newStatus ? callback() : undefined);
    },

    onSuccess(callback) {
      return this.onStatusChangeTo('successful', callback);
    },

    onFailure(callback) {
      return this.onStatusChangeTo('failed', callback);
    },

    onCancel(callback) {
      return this.onStatusChangeTo('canceled', callback);
    }

  };
};

exports.useTask = useTask;
//# sourceMappingURL=useTask.js.map
