import { defineProperty as _defineProperty } from '../../_virtual/_rollupPluginBabelHelpers.js';
import NodeBundlr, { WebBundlr } from '@bundlr-network/client';
import BigNumber from 'bignumber.js';
import { AssetUploadFailedError, BundlrWithdrawError, FailedToConnectToBundlrAddressError, FailedToInitializeBundlrError } from '../../errors/BundlrError.js';
import { KeypairIdentityDriver } from '../keypairIdentity/KeypairIdentityDriver.js';
import { StorageDriver } from '../../types/StorageDriver.js';
import { SolAmount } from '../../utils/SolAmount.js';

class BundlrStorageDriver extends StorageDriver {
  constructor(metaplex, options = {}) {
    var _options$withdrawAfte;

    super(metaplex);

    _defineProperty(this, "bundlr", null);

    this.options = {
      providerUrl: metaplex.connection.rpcEndpoint,
      ...options
    };
    this._withdrawAfterUploading = (_options$withdrawAfte = options.withdrawAfterUploading) !== null && _options$withdrawAfte !== void 0 ? _options$withdrawAfte : true;
  }

  async getBalance() {
    const bundlr = await this.getBundlr();
    const balance = await bundlr.getLoadedBalance();
    return SolAmount.fromLamports(balance);
  }

  async getPrice(...files) {
    const price = await this.getMultipliedPrice(this.getBytes(files));
    return SolAmount.fromLamports(price);
  }

  async upload(file) {
    const [uri] = await this.uploadAll([file]);
    return uri;
  }

  async uploadAll(files) {
    await this.fund(files);
    const promises = files.map(file => this.uploadFile(file));
    const uris = await Promise.all(promises);

    if (this.shouldWithdrawAfterUploading()) {
      await this.withdrawAll();
    }

    return uris;
  }

  async fundingNeeded(filesOrBytes, skipBalanceCheck = false) {
    const price = await this.getMultipliedPrice(this.getBytes(filesOrBytes));

    if (skipBalanceCheck) {
      return price;
    }

    const bundlr = await this.getBundlr();
    const balance = await bundlr.getLoadedBalance();
    return price.isGreaterThan(balance) ? price.minus(balance) : new BigNumber(0);
  }

  async needsFunding(filesOrBytes, skipBalanceCheck = false) {
    const fundingNeeded = await this.fundingNeeded(filesOrBytes, skipBalanceCheck);
    return fundingNeeded.isGreaterThan(0);
  }

  async fund(filesOrBytes, skipBalanceCheck = false) {
    const bundlr = await this.getBundlr();
    const fundingNeeded = await this.fundingNeeded(filesOrBytes, skipBalanceCheck);

    if (!fundingNeeded.isGreaterThan(0)) {
      return;
    } // TODO: Catch errors and wrap in BundlrErrors.


    await bundlr.fund(fundingNeeded);
  }

  getBytes(filesOrBytes) {
    if (typeof filesOrBytes === 'number') {
      return filesOrBytes;
    }

    return filesOrBytes.reduce((total, file) => total + file.getBytes(), 0);
  }

  async getMultipliedPrice(bytes) {
    var _this$options$priceMu;

    const bundlr = await this.getBundlr();
    const price = await bundlr.getPrice(bytes);
    return price.multipliedBy((_this$options$priceMu = this.options.priceMultiplier) !== null && _this$options$priceMu !== void 0 ? _this$options$priceMu : 1.5).decimalPlaces(0);
  }

  async uploadFile(file) {
    const bundlr = await this.getBundlr();
    const {
      status,
      data
    } = await bundlr.uploader.upload(file.toBuffer(), file.getTagsWithContentType());

    if (status >= 300) {
      throw new AssetUploadFailedError(status);
    }

    return `https://arweave.net/${data.id}`;
  }

  async withdrawAll() {
    // TODO(loris): Replace with "withdrawAll" when available on Bundlr.
    const balance = await this.getBalance();
    const minimumBalance = SolAmount.fromLamports(5000);

    if (balance.isLessThan(minimumBalance)) {
      return;
    }

    const balanceToWithdraw = balance.minus(minimumBalance).getLamports();
    await this.withdraw(balanceToWithdraw);
  }

  async withdraw(lamports) {
    const bundlr = await this.getBundlr();
    const {
      status
    } = await bundlr.withdrawBalance(new BigNumber(lamports));

    if (status >= 300) {
      throw new BundlrWithdrawError(status);
    }
  }

  shouldWithdrawAfterUploading() {
    return this._withdrawAfterUploading;
  }

  withdrawAfterUploading() {
    this._withdrawAfterUploading = true;
    return this;
  }

  dontWithdrawAfterUploading() {
    this._withdrawAfterUploading = false;
    return this;
  }

  async getBundlr() {
    var _this$options$address, _this$options;

    if (this.bundlr) return this.bundlr;
    const currency = 'solana';
    const address = (_this$options$address = (_this$options = this.options) === null || _this$options === void 0 ? void 0 : _this$options.address) !== null && _this$options$address !== void 0 ? _this$options$address : 'https://node1.bundlr.network';
    const options = {
      timeout: this.options.timeout,
      providerUrl: this.options.providerUrl
    };
    const identity = this.metaplex.identity();
    const bundlr = identity instanceof KeypairIdentityDriver ? new NodeBundlr(address, currency, identity.keypair.secretKey, options) : new WebBundlr(address, currency, identity, options);

    try {
      // Check for valid bundlr node.
      await bundlr.utils.getBundlerAddress(currency);
    } catch (error) {
      throw new FailedToConnectToBundlrAddressError(address, error);
    }

    if (bundlr instanceof WebBundlr) {
      try {
        // Try to initiate bundlr.
        await bundlr.ready();
      } catch (error) {
        throw new FailedToInitializeBundlrError(error);
      }
    }

    this.bundlr = bundlr;
    return bundlr;
  }

}

export { BundlrStorageDriver };
//# sourceMappingURL=BundlrStorageDriver.js.map
