import { defineProperty as _defineProperty } from '../_virtual/_rollupPluginBabelHelpers.js';
import EventEmitter from 'eventemitter3';

class Plan {
  constructor(plan) {
    var _plan$steps, _plan$eventEmitter;

    _defineProperty(this, "executing", false);

    _defineProperty(this, "executed", false);

    _defineProperty(this, "failed", false);

    this.promise = plan.promise;
    this.steps = (_plan$steps = plan.steps) !== null && _plan$steps !== void 0 ? _plan$steps : [];
    this.eventEmitter = (_plan$eventEmitter = plan.eventEmitter) !== null && _plan$eventEmitter !== void 0 ? _plan$eventEmitter : new EventEmitter();
  }

  static make() {
    return new Plan({
      promise: async initialState => initialState
    });
  }

  addStep(stepOrName, stepHandler, stepOptions) {
    const {
      newStep,
      handler
    } = Plan.parseInputStep(stepOrName, stepHandler, stepOptions);

    const promise = async (initialState, plan) => {
      let state;

      try {
        state = await this.promise(initialState, plan);
      } catch (error) {
        Plan.changeStepStatus(plan, newStep, 'canceled');
        throw error;
      }

      return Plan.processStep(plan, state, newStep, handler);
    };

    return new Plan({
      promise,
      steps: [...this.steps, newStep],
      eventEmitter: this.eventEmitter
    });
  }

  prependStep(stepOrName, stepHandler, stepOptions) {
    const {
      newStep,
      handler
    } = Plan.parseInputStep(stepOrName, stepHandler, stepOptions);

    const promise = async (newInitialState, plan) => {
      let initialState;

      try {
        initialState = await Plan.processStep(plan, newInitialState, newStep, handler);
      } catch (error) {
        this.steps.forEach(step => Plan.changeStepStatus(plan, step, 'canceled'));
        throw error;
      }

      return this.promise(initialState, plan);
    };

    return new Plan({
      promise,
      steps: [newStep, ...this.steps],
      eventEmitter: this.eventEmitter
    });
  }

  onChange(listener) {
    this.eventEmitter.addListener('change', listener);
    return this;
  }

  getSteps() {
    return this.steps;
  }

  getVisibleSteps() {
    return this.steps.filter(step => !step.hidden);
  }

  async execute(initialState) {
    try {
      this.executing = true;
      this.executed = false;
      this.failed = false;
      const state = initialState !== null && initialState !== void 0 ? initialState : undefined;
      return await this.promise(state, this);
    } catch (error) {
      this.failed = true;
      throw error;
    } finally {
      this.executing = false;
      this.executed = true;
    }
  }

  static parseInputStep(stepOrName, stepHandler, stepOptions) {
    var _inputStep$hidden, _inputStep$optional;

    let inputStep;

    if (typeof stepOrName === 'string') {
      if (!stepHandler) {
        throw new TypeError('Missing step handler');
      }

      inputStep = {
        name: stepOrName,
        handler: stepHandler,
        ...stepOptions
      };
    } else {
      inputStep = stepOrName;
    }

    const {
      handler
    } = inputStep;
    const newStep = {
      name: inputStep.name,
      status: 'pending',
      hidden: (_inputStep$hidden = inputStep.hidden) !== null && _inputStep$hidden !== void 0 ? _inputStep$hidden : false,
      optional: (_inputStep$optional = inputStep.optional) !== null && _inputStep$optional !== void 0 ? _inputStep$optional : false,
      onError: inputStep.onError
    };
    return {
      newStep,
      handler
    };
  }

  static async processStep(plan, from, step, handler) {
    this.changeStepStatus(plan, step, 'running');

    try {
      const to = await handler(from, plan);
      this.changeStepStatus(plan, step, 'successful');
      return to;
    } catch (error) {
      var _step$onError;

      this.changeStepStatus(plan, step, 'failed');
      (_step$onError = step.onError) === null || _step$onError === void 0 ? void 0 : _step$onError.call(step, error);

      if (step.optional) {
        // If a step is optional, it's destination state should match
        // the source state. Otherwise, steps cannot be composed.
        return from;
      } else {
        throw error;
      }
    }
  }

  static notifyChange(plan, step) {
    plan.eventEmitter.emit('change', step, plan);
  }

  static changeStepStatus(plan, step, newStatus) {
    step.status = newStatus;
    this.notifyChange(plan, step);
  }

}

export { Plan };
//# sourceMappingURL=Plan.js.map
